//========================================================================
// GLFW - An OpenGL framework
// File:     input.c
// Platform: Win32 & X11 (Unix)
// Version:  2.3
// Date:     2002.11.23
// Author:   Marcus Geelnard (marcus.geelnard@home.se)
// WWW:      http://hem.passagen.se/opengl/glfw/
//------------------------------------------------------------------------
// Copyright (c) 2002 Marcus Geelnard
//
// GLFW is provided "as-is", without any express or implied warranty,
// without even the implied warranty of fitness for a particular purpose.
// In no event will the author be held liable for any damage araising from
// the use of GLFW.
//
// Permission is granted to anyone to use GLFW for any purpose, including
// commercial applications, and to alter it and redistribute it freely,
// subject to the following restrictions:
//
// 1. You must not claim that you wrote GLFW.
// 2. If you modify any parts of GLFW, it must be clearly stated which
//    modifications have been made, and that those modifications are not
//    the work of the author of GLFW.
// 3. If you use GLFW in a product, an acknowledgment in the product
//    documentation would be appreciated, but is not required.
// 4. This notice may not be removed or altered from any source
//    distribution of GLFW.
//
// Marcus Geelnard
// marcus.geelnard@home.se
//========================================================================

#include "internal.h"


//========================================================================
// glfwGetKey()
//========================================================================

GLFWAPI int GLFWAPIENTRY glfwGetKey( int key )
{
    // Is GLFW initialized?
    if( !_glfwInitialised || !_glfwWin.Opened )
    {
        return GLFW_RELEASE;
    }

    // Is it a valid key?
    if( key < 0 || key > GLFW_KEY_LAST )
    {
        return GLFW_RELEASE;
    }

    if( _glfwInput.Key[ key ] == 2 )
    {
        // Sticky mode: release key now
        _glfwInput.Key[ key ] = 0;
        return GLFW_PRESS;
    }

    return (int) _glfwInput.Key[ key ];
}


//========================================================================
// glfwGetMouseButton()
//========================================================================

GLFWAPI int GLFWAPIENTRY glfwGetMouseButton( int button )
{
    // Is GLFW initialized?
    if( !_glfwInitialised || !_glfwWin.Opened )
    {
        return GLFW_RELEASE;
    }

    // Is it a valid mouse button?
    if( button < 0 || button > GLFW_MOUSE_BUTTON_LAST )
    {
        return GLFW_RELEASE;
    }

    if( _glfwInput.MouseButton[ button ] == 2 )
    {
        // Sticky mode: release mouse button now
        _glfwInput.MouseButton[ button ] = 0;
        return GLFW_PRESS;
    }

    return (int) _glfwInput.MouseButton[ button ];
}


//========================================================================
// glfwGetMousePos()
//========================================================================

GLFWAPI void GLFWAPIENTRY glfwGetMousePos( int *xpos, int *ypos )
{
    // Is GLFW initialized?
    if( !_glfwInitialised || !_glfwWin.Opened )
    {
        return;
    }

    // Return mouse position
    if( xpos != NULL )
    {
        *xpos = _glfwInput.MousePosX;
    }
    if( ypos != NULL )
    {
        *ypos = _glfwInput.MousePosY;
    }
}


//========================================================================
// glfwSetMousePos()
//========================================================================

GLFWAPI void GLFWAPIENTRY glfwSetMousePos( int xpos, int ypos )
{
    // Is GLFW initialized?
    if( !_glfwInitialised || !_glfwWin.Opened )
    {
        return;
    }

    // Set GLFW mouse position
    _glfwInput.MousePosX = xpos;
    _glfwInput.MousePosY = ypos;

    // If we have a locked mouse, physical position is the center of the
    // window
    if( _glfwWin.MouseLock )
    {
        xpos = _glfwWin.Width >> 1;
        ypos = _glfwWin.Height >> 1;
    }

    // Update cursor position
#ifdef _WIN32
    {
        POINT pos;

        // Convert mouse coordinates to screen coordinates
        pos.x = xpos;
        pos.y = ypos;
        ClientToScreen( _glfwWin.Wnd, &pos );

        // Set cursor position
        SetCursorPos( pos.x, pos.y );
    }
#else
    XWarpPointer( _glfwWin.Dpy, None, _glfwWin.Win, 0,0,0,0, xpos, ypos );
#endif // _WIN32
}


//========================================================================
// glfwGetMouseWheel()
//========================================================================

GLFWAPI int GLFWAPIENTRY glfwGetMouseWheel( void )
{
    // Is GLFW initialized?
    if( !_glfwInitialised || !_glfwWin.Opened )
    {
        return 0;
    }

    // Return mouse wheel position
    return _glfwInput.WheelPos;
}


//========================================================================
// glfwSetMouseWheel()
//========================================================================

GLFWAPI void GLFWAPIENTRY glfwSetMouseWheel( int pos )
{
    // Is GLFW initialized?
    if( !_glfwInitialised || !_glfwWin.Opened )
    {
        return;
    }

    // Set mouse wheel position
    _glfwInput.WheelPos = pos;
}


//========================================================================
// glfwSetKeyCallback() - Set callback function for keyboard input
//========================================================================

GLFWAPI void GLFWAPIENTRY glfwSetKeyCallback( GLFWkeyfun cbfun )
{
    // Is GLFW initialized?
    if( !_glfwInitialised || !_glfwWin.Opened )
    {
        return;
    }

    // Set callback function
    _glfwKeyCallback = cbfun;
}


//========================================================================
// glfwSetMouseButtonCallback() - Set callback function for mouse clicks
//========================================================================

GLFWAPI void GLFWAPIENTRY glfwSetMouseButtonCallback( GLFWmousebuttonfun cbfun )
{
    // Is GLFW initialized?
    if( !_glfwInitialised || !_glfwWin.Opened )
    {
        return;
    }

    // Set callback function
    _glfwMouseButtonCallback = cbfun;
}


//========================================================================
// glfwSetMousePosCallback() - Set callback function for mouse moves
//========================================================================

GLFWAPI void GLFWAPIENTRY glfwSetMousePosCallback( GLFWmouseposfun cbfun )
{
    // Is GLFW initialized?
    if( !_glfwInitialised || !_glfwWin.Opened )
    {
        return;
    }

    // Set callback function
    _glfwMousePosCallback = cbfun;
}


//========================================================================
// glfwSetMouseWheelCallback() - Set callback function for mouse wheel
//========================================================================

GLFWAPI void GLFWAPIENTRY glfwSetMouseWheelCallback( GLFWmousewheelfun cbfun )
{
    // Is GLFW initialized?
    if( !_glfwInitialised || !_glfwWin.Opened )
    {
        return;
    }

    // Set callback function
    _glfwMouseWheelCallback = cbfun;
}
