//========================================================================
// GLFW - An OpenGL framework
// File:     window.c
// Platform: X11 (Unix)
// Version:  2.3
// Date:     2002.11.23
// Author:   Marcus Geelnard (marcus.geelnard@home.se)
// WWW:      http://hem.passagen.se/opengl/glfw/
//------------------------------------------------------------------------
// Copyright (c) 2002 Marcus Geelnard
//
// GLFW is provided "as-is", without any express or implied warranty,
// without even the implied warranty of fitness for a particular purpose.
// In no event will the author be held liable for any damage araising from
// the use of GLFW.
//
// Permission is granted to anyone to use GLFW for any purpose, including
// commercial applications, and to alter it and redistribute it freely,
// subject to the following restrictions:
//
// 1. You must not claim that you wrote GLFW.
// 2. If you modify any parts of GLFW, it must be clearly stated which
//    modifications have been made, and that those modifications are not
//    the work of the author of GLFW.
// 3. If you use GLFW in a product, an acknowledgment in the product
//    documentation would be appreciated, but is not required.
// 4. This notice may not be removed or altered from any source
//    distribution of GLFW.
//
// Marcus Geelnard
// marcus.geelnard@home.se
//========================================================================

#include "internal.h"


//************************************************************************
//****                  GLFW internal functions                       ****
//************************************************************************

//========================================================================
// _glfwChooseVisual() - We do our own function here, since
// glXChooseVisual does not behave as we want it to (not according to the
// GLFW specs)
//========================================================================

XVisualInfo * _glfwChooseVisual( Display *Dpy, int Screen, int r, int g,
    int b, int a, int d, int s, int ar, int ag, int ab, int aa, int aux,
    int stereo )
{
    XVisualInfo *VI, *VI_list, VI_tmp;
    int  nitems_return, i;
    int  vi_gl, vi_rgba, vi_double, vi_stereo;
    int  vi_r, vi_g, vi_b, vi_a, vi_d, vi_s, vi_ar, vi_ag, vi_ab, vi_aa;
    int  vi_aux;
    int  color, accum, vi_accum;
    int  missing, color_diff, extra_diff;
    int  best_vis, best_missing, best_color_diff, best_extra_diff;

    // Get list of visuals for this screen & display
    VI_tmp.screen = Screen;
    VI_list = XGetVisualInfo( Dpy, VisualScreenMask, &VI_tmp,
                              &nitems_return );
    if( VI_list == NULL )
    {
        return NULL;
    }

    // Pick some prefered color depth if the user did not request a
    // specific depth (note: if the user did not request a specific color
    // depth, this will not be a driving demand, it's only here to avoid
    // selection randomness)
    color = (r > 0 || g > 0 || b > 0);
    if( !color )
    {
        r = g = b = 8;
    }

    // Make sure that stereo is 1 or 0
    stereo = stereo ? 1 : 0;

    // Convenience pre-calculation
    accum = (ar > 0 || ag > 0 || ab > 0 || aa > 0);

    // Loop through list of visuals to find best match
    best_vis        = -1;
    best_missing    = 0x7fffffff;
    best_color_diff = 0x7fffffff;
    best_extra_diff = 0x7fffffff;
    for( i = 0; i < nitems_return; i ++ )
    {
        // We want GL, RGBA & DOUBLEBUFFER, and NOT STEREO / STEREO
        glXGetConfig( Dpy, &VI_list[i], GLX_USE_GL, &vi_gl );
        glXGetConfig( Dpy, &VI_list[i], GLX_RGBA, &vi_rgba );
        glXGetConfig( Dpy, &VI_list[i], GLX_DOUBLEBUFFER, &vi_double );
        glXGetConfig( Dpy, &VI_list[i], GLX_STEREO, &vi_stereo );
        vi_stereo = vi_stereo ? 1 : 0;
        if( vi_gl && vi_rgba && vi_double && (vi_stereo == stereo) )
        {
            // Get visual color parameters
            glXGetConfig( Dpy, &VI_list[i], GLX_RED_SIZE, &vi_r );
            glXGetConfig( Dpy, &VI_list[i], GLX_GREEN_SIZE, &vi_g );
            glXGetConfig( Dpy, &VI_list[i], GLX_BLUE_SIZE, &vi_b );

            // Get visual "extra" parameters
            glXGetConfig( Dpy, &VI_list[i], GLX_ALPHA_SIZE, &vi_a );
            glXGetConfig( Dpy, &VI_list[i], GLX_DEPTH_SIZE, &vi_d );
            glXGetConfig( Dpy, &VI_list[i], GLX_STENCIL_SIZE, &vi_s );
            glXGetConfig( Dpy, &VI_list[i], GLX_ACCUM_RED_SIZE, &vi_ar );
            glXGetConfig( Dpy, &VI_list[i], GLX_ACCUM_GREEN_SIZE, &vi_ag );
            glXGetConfig( Dpy, &VI_list[i], GLX_ACCUM_BLUE_SIZE, &vi_ab );
            glXGetConfig( Dpy, &VI_list[i], GLX_ACCUM_ALPHA_SIZE, &vi_aa );
            glXGetConfig( Dpy, &VI_list[i], GLX_AUX_BUFFERS, &vi_aux );
            vi_accum = (vi_ar > 0 || vi_ag > 0 || vi_ab > 0 || vi_aa > 0);

            // Check how many buffers are missing
            missing = 0;
            if( a > 0 && vi_a == 0 ) missing ++;
            if( d > 0 && vi_d == 0 ) missing ++;
            if( s > 0 && vi_s == 0 ) missing ++;
            if( accum && !vi_accum ) missing ++;
            if( aux > 0 && vi_aux == 0 ) missing ++;

            // Calculate color diff
            color_diff = (r - vi_r) * (r - vi_r) +
                         (g - vi_g) * (g - vi_g) +
                         (b - vi_b) * (b - vi_b);

            // Calculate "extra" diff
            extra_diff = 0;
            if( a > 0 )
            {
                extra_diff += (a - vi_a) * (a - vi_a);
            }
            if( d > 0 )
            {
                extra_diff += (d - vi_d) * (d - vi_d);
            }
            if( s > 0 )
            {
                extra_diff += (s - vi_s) * (s - vi_s);
            }
            if( accum )
            {
                extra_diff += (ar - vi_ar) * (ar - vi_ar) +
                              (ag - vi_ag) * (ag - vi_ag) +
                              (ab - vi_ab) * (ab - vi_ab) +
                              (aa - vi_aa) * (aa - vi_aa);
            }
            if( aux > 0 )
            {
                extra_diff += (aux - vi_aux) * (aux - vi_aux);
            }

            // Check if this is a better match. We implement some
            // complicated rules, by prioritizing in this order:
            //  1) Visuals with the least number of missing buffers always
            //     have priority
            //  2a) If (r,g,b)!=(0,0,0), color depth has priority over
            //      other buffers
            //  2b) If (r,g,b)==(0,0,0), other buffers have priority over
            //      color depth
            if( missing < best_missing )
            {
                best_vis = i;
            }
            else if( missing == best_missing )
            {
                if( color )
                {
                    if( (color_diff < best_color_diff) ||
                        (color_diff == best_color_diff &&
                         extra_diff < best_extra_diff) )
                    {
                        best_vis = i;
                    }
                }
                else
                {
                    if( (extra_diff < best_extra_diff) ||
                        (extra_diff == best_extra_diff &&
                         color_diff < best_color_diff) )
                    {
                        best_vis = i;
                    }
                }
            }
            if( best_vis == i )
            {
                best_missing    = missing;
                best_color_diff = color_diff;
                best_extra_diff = extra_diff;
            }
        }
    }

    // Copy best visual to a visual to return
    if( best_vis >= 0 )
    {
        VI = XGetVisualInfo( Dpy, VisualIDMask, &VI_list[ best_vis ],
                             &nitems_return );
    }
    else
    {
        VI = NULL;
    }

    // Free visuals list
    XFree( VI_list );

    return VI;
}


//========================================================================
// _glfwWaitForNotify()
//========================================================================

Bool _glfwWaitForNotify( Display *d, XEvent *e, char *arg )
{
    return (e->type == MapNotify) && (e->xmap.window == (Window)arg);
}


//========================================================================
// _glfwClearInput() - Clear all input events
//========================================================================

static void _glfwClearInput( void )
{
    int i;

    // Release all keyboard keys
    for( i = 0; i <= GLFW_KEY_LAST; i ++ )
    {
        _glfwInput.Key[ i ] = 0;
    }

    // Release all mouse buttons
    for( i = 0; i <= GLFW_MOUSE_BUTTON_LAST; i ++ )
    {
        _glfwInput.MouseButton[ i ] = 0;
    }

    // Set mouse position to (0,0)
    _glfwInput.MousePosX = 0;
    _glfwInput.MousePosY = 0;

    // Set mouse wheel position to 0
    _glfwInput.WheelPos = 0;

    // The default is to use non sticky keys and mouse buttons
    _glfwInput.StickyKeys = GL_FALSE;
    _glfwInput.StickyMouseButtons = GL_FALSE;

    // The default is to disable key repeat
    _glfwInput.KeyRepeat = GL_FALSE;
}


//========================================================================
// _glfwInputKey() - Register keyboard activity
//========================================================================

static void _glfwInputKey( int key, int action )
{
    int keyrepeat = 0;

    if( key >= 0 && key <= GLFW_KEY_LAST )
    {
        // Are we trying to release an already released key?
        if( action == GLFW_RELEASE && _glfwInput.Key[ key ] != 1 )
        {
            return;
        }

        // Register key action
        if( action == GLFW_RELEASE && _glfwInput.StickyKeys )
        {
            _glfwInput.Key[ key ] = 2;
        }
        else
        {
            keyrepeat = (_glfwInput.Key[ key ] == GLFW_PRESS) &&
                        (action == GLFW_PRESS);
            _glfwInput.Key[ key ] = action;
        }

        // Call user callback function
        if( _glfwKeyCallback && (_glfwInput.KeyRepeat || !keyrepeat) )
        {
            _glfwKeyCallback( key, action );
        }
    }
}


//========================================================================
// _glfwInputMouseClick() - Register mouse button clicks
//========================================================================

static void _glfwInputMouseClick( int button, int action )
{
    if( button >= 0 && button <= GLFW_MOUSE_BUTTON_LAST )
    {
        // Register mouse button action
        if( action == GLFW_RELEASE && _glfwInput.StickyMouseButtons )
        {
            _glfwInput.MouseButton[ button ] = 2;
        }
        else
        {
            _glfwInput.MouseButton[ button ] = action;
        }

        // Call user callback function
        if( _glfwMouseButtonCallback )
        {
            _glfwMouseButtonCallback( button, action );
        }
    }
}


//========================================================================
// _glfwTranslateKey() - Translates a X Window key to internal coding
//========================================================================

static int _glfwTranslateKey( int keycode )
{
    int key;

    // Try secondary keysym, for numeric keypad keys
    // Note: This way we always force "NumLock = ON", which at least
    // enables GLFW users to detect numeric keypad keys
    key = XKeycodeToKeysym( _glfwWin.Dpy, keycode, 1 );
    switch( key )
    {
        // Numeric keypad
        case XK_KP_0:         return GLFW_KEY_KP_0;
        case XK_KP_1:         return GLFW_KEY_KP_1;
        case XK_KP_2:         return GLFW_KEY_KP_2;
        case XK_KP_3:         return GLFW_KEY_KP_3;
        case XK_KP_4:         return GLFW_KEY_KP_4;
        case XK_KP_5:         return GLFW_KEY_KP_5;
        case XK_KP_6:         return GLFW_KEY_KP_6;
        case XK_KP_7:         return GLFW_KEY_KP_7;
        case XK_KP_8:         return GLFW_KEY_KP_8;
        case XK_KP_9:         return GLFW_KEY_KP_9;
        case XK_KP_Separator:
        case XK_KP_Decimal:   return GLFW_KEY_KP_DECIMAL;
        case XK_KP_Equal:
        case XK_KP_Enter:     return GLFW_KEY_KP_ENTER;
        default:              break;
    }

    // Now try pimary keysym
    key = XKeycodeToKeysym( _glfwWin.Dpy, keycode, 0 );
    switch( key )
    {
        // Special keys (non character keys)
        case XK_Escape:       return GLFW_KEY_ESC;
        case XK_Tab:          return GLFW_KEY_TAB;
        case XK_Shift_L:      return GLFW_KEY_LSHIFT;
        case XK_Shift_R:      return GLFW_KEY_RSHIFT;
        case XK_Control_L:    return GLFW_KEY_LCTRL;
        case XK_Control_R:    return GLFW_KEY_RCTRL;
        case XK_Meta_L:
        case XK_Alt_L:        return GLFW_KEY_LALT;
        case XK_Mode_switch:  // Mapped to Alt_R on many keyboards
        case XK_Meta_R:
        case XK_Alt_R:        return GLFW_KEY_RALT;
        case XK_KP_Delete:
        case XK_Delete:       return GLFW_KEY_DEL;
        case XK_BackSpace:    return GLFW_KEY_BACKSPACE;
        case XK_Return:       return GLFW_KEY_ENTER;
        case XK_KP_Home:
        case XK_Home:         return GLFW_KEY_HOME;
        case XK_KP_End:
        case XK_End:          return GLFW_KEY_END;
        case XK_KP_Page_Up:
        case XK_Page_Up:      return GLFW_KEY_PAGEUP;
        case XK_KP_Page_Down:
        case XK_Page_Down:    return GLFW_KEY_PAGEDOWN;
        case XK_KP_Insert:
        case XK_Insert:       return GLFW_KEY_INSERT;
        case XK_KP_Left:
        case XK_Left:         return GLFW_KEY_LEFT;
        case XK_KP_Right:
        case XK_Right:        return GLFW_KEY_RIGHT;
        case XK_KP_Down:
        case XK_Down:         return GLFW_KEY_DOWN;
        case XK_KP_Up:
        case XK_Up:           return GLFW_KEY_UP;
        case XK_F1:           return GLFW_KEY_F1;
        case XK_F2:           return GLFW_KEY_F2;
        case XK_F3:           return GLFW_KEY_F3;
        case XK_F4:           return GLFW_KEY_F4;
        case XK_F5:           return GLFW_KEY_F5;
        case XK_F6:           return GLFW_KEY_F6;
        case XK_F7:           return GLFW_KEY_F7;
        case XK_F8:           return GLFW_KEY_F8;
        case XK_F9:           return GLFW_KEY_F9;
        case XK_F10:          return GLFW_KEY_F10;
        case XK_F11:          return GLFW_KEY_F11;
        case XK_F12:          return GLFW_KEY_F12;
        case XK_equal:        return (int) '=';
        case XK_minus:        return (int) '-';

        // Numeric keypad
        case XK_KP_Divide:    return GLFW_KEY_KP_DIVIDE;
        case XK_KP_Multiply:  return GLFW_KEY_KP_MULTIPLY;
        case XK_KP_Subtract:  return GLFW_KEY_KP_SUBTRACT;
        case XK_KP_Add:       return GLFW_KEY_KP_ADD;
        case XK_KP_Equal:
        case XK_KP_Enter:     return GLFW_KEY_KP_ENTER;

        // The rest (should be alpha numeric keys)
        default:
            if( key >= XK_space && key <= XK_asciitilde )
            {
                key = key - XK_space + ' ';
            }
            if( key >= 'a' && key <= 'z' )
            {
                key = key - 'a' + 'A';
            }
            if( key >= ' ' && key <= '~' )
            {
                return key;
            }

            // There is one more possibility: Check for "undocumented" non
            // alpha-numeric keys (e.g. '[' etc). Use UK keyboard layout
            // for these, and use the keycode directly (not nice, but
            // should work on most X11 stations). Problems:
            //  1) Keycodes may differ between systems
            //  2) We may end up with two keys having the same
            //     interpretationt on international keyboards
            // Anyway, this solved things for my Swedish keyboard on
            // XFree86, so I'm happy. Let me know if you're not...
            switch( keycode )
            {
                case 49:              return (int)'`';
                case 20:              return (int)'-';
                case 21:              return (int)'=';
                case 34:              return (int)'[';
                case 35:              return (int)']';
                case 47:              return (int)';';
                case 48:              return (int)'\'';
                case 51:              return (int)'#';
                case 94:              return (int)'\\';
                case 59:              return (int)',';
                case 60:              return (int)'.';
                case 61:              return (int)'/';
                default:              break;
            }

            return GLFW_KEY_UNKNOWN;
    }
}


//========================================================================
// _glfwGetNextEvent() - Get next X event (called by glfwPollEvents)
//========================================================================

static int _glfwGetNextEvent( void )
{
    XEvent event, next_event;
    XSetWindowAttributes wa;

    XNextEvent( _glfwWin.Dpy, &event );

    // Handle certain window messages
    switch( event.type )
    {
        // Is a key being pressed?
        case KeyPress:
            _glfwInputKey( _glfwTranslateKey( event.xkey.keycode ),
                           GLFW_PRESS );
            break;

        // Is a key or being released?
        case KeyRelease:
            // Do not report key releases for key repeats. For key repeats
            // we will get KeyRelease/KeyPress pairs with identical time
            // stamps. User selected key repeat filtering is handled in
            // _glfwInputKey().
            if( XEventsQueued( _glfwWin.Dpy, QueuedAfterReading ) )
            {
                XPeekEvent( _glfwWin.Dpy, &next_event );
                if( next_event.type == KeyPress &&
                    next_event.xkey.window == event.xkey.window &&
                    next_event.xkey.keycode == event.xkey.keycode &&
                    next_event.xkey.time == event.xkey.time )
                {
                    // Do not report anything for this event
                    break;
                }
            }
            _glfwInputKey( _glfwTranslateKey( event.xkey.keycode ),
                           GLFW_RELEASE );
            break;

        // Were any of the mouse-buttons pressed?
        case ButtonPress:
            if( event.xbutton.state & Button1Mask ||
                event.xbutton.button == Button1 )
            {
                _glfwInputMouseClick( GLFW_MOUSE_BUTTON_LEFT,
                                      GLFW_PRESS );
            }
            if( event.xbutton.state & Button2Mask ||
                event.xbutton.button == Button2 )
            {
                _glfwInputMouseClick( GLFW_MOUSE_BUTTON_MIDDLE,
                                      GLFW_PRESS );
            }
            if( event.xbutton.state & Button3Mask ||
                event.xbutton.button == Button3 )
            {
                _glfwInputMouseClick( GLFW_MOUSE_BUTTON_RIGHT,
                                      GLFW_PRESS );
            }

            // XFree86 3.3.2 and later translates mouse wheel up/down into
            // mouse button 4 & 5 presses
            if( event.xbutton.state & Button4Mask ||
                event.xbutton.button == Button4 )
            {
                _glfwInput.WheelPos ++;  // To verify: is this up or down?
                if( _glfwMouseWheelCallback )
                {
                    _glfwMouseWheelCallback( _glfwInput.WheelPos );
                }
            }
            if( event.xbutton.state & Button5Mask ||
                event.xbutton.button == Button5 )
            {
                _glfwInput.WheelPos --;
                if( _glfwMouseWheelCallback )
                {
                    _glfwMouseWheelCallback( _glfwInput.WheelPos );
                }
            }
            break;

        // Were any of the mouse-buttons released?
        case ButtonRelease:
            if( event.xbutton.state & Button1Mask ||
                event.xbutton.button == Button1 )
            {
                _glfwInputMouseClick( GLFW_MOUSE_BUTTON_LEFT,
                                      GLFW_RELEASE );
            }
            if( event.xbutton.state & Button2Mask ||
                event.xbutton.button == Button2 )
            {
                _glfwInputMouseClick( GLFW_MOUSE_BUTTON_MIDDLE,
                                      GLFW_RELEASE );
            }
            if( event.xbutton.state & Button3Mask ||
                event.xbutton.button == Button3 )
            {
                _glfwInputMouseClick( GLFW_MOUSE_BUTTON_RIGHT,
                                      GLFW_RELEASE );
            }
            break;

        // Was the mouse moved?
        case MotionNotify:
            if( event.xmotion.x != _glfwInput.OldMouseX ||
                event.xmotion.y != _glfwInput.OldMouseY )
            {
                if( _glfwWin.MouseLock )
                {
                    _glfwInput.MousePosX += event.xmotion.x -
                                            _glfwInput.OldMouseX;
                    _glfwInput.MousePosY += event.xmotion.y -
                                            _glfwInput.OldMouseY;
                }
                else
                {
                    _glfwInput.MousePosX = event.xmotion.x;
                    _glfwInput.MousePosY = event.xmotion.y;
                }
                _glfwInput.OldMouseX = event.xmotion.x;
                _glfwInput.OldMouseY = event.xmotion.y;
                _glfwInput.MouseMoved = GL_TRUE;

                // Call user callback function
                if( _glfwMousePosCallback )
                {
                    _glfwMousePosCallback( _glfwInput.MousePosX,
                                           _glfwInput.MousePosY );
                }
            }
            break;

        // Was the window resized?
        case ConfigureNotify:
            if( event.xconfigure.width != _glfwWin.Width ||
                event.xconfigure.height != _glfwWin.Height )
            {
                _glfwWin.Width = event.xconfigure.width;
                _glfwWin.Height = event.xconfigure.height;
                if( _glfwWindowSizeCallback )
                {
                    _glfwWindowSizeCallback( _glfwWin.Width,
                                             _glfwWin.Height );
                }
            }
            break;

        // Was the window closed by the window manager?
        case ClientMessage:
            if( event.xclient.data.l[ 0 ] == _glfwWin.WMDeleteWindow )
            {
                return GL_TRUE;
            }
            break;

        // Was the window iconified?
        case UnmapNotify:
            // Show mouse pointer
            if( _glfwWin.PointerHidden )
            {
                XUndefineCursor( _glfwWin.Dpy, _glfwWin.Win );
                _glfwWin.PointerHidden = GL_FALSE;
            }

            // Un-grab mouse pointer
            if( _glfwWin.PointerGrabbed )
            {
                XUngrabPointer( _glfwWin.Dpy, CurrentTime );
                _glfwWin.PointerGrabbed = GL_FALSE;
            }

            _glfwWin.Iconified = GL_TRUE;
            break;

        // Was the window un-iconified?
        case MapNotify:
            if( _glfwWin.Iconified )
            {
                // Lock mouse, if necessary
                if( _glfwWin.MouseLock )
                {
                    // Hide cursor
                    if( !_glfwWin.PointerHidden )
                    {
                        XDefineCursor( _glfwWin.Dpy, _glfwWin.Win,
                            _glfwCreateNULLCursor( _glfwWin.Dpy,
                                                   _glfwWin.Win ) );
                        _glfwWin.PointerHidden = GL_TRUE;
                    }

                    // Grab cursor
                    if( !_glfwWin.PointerGrabbed )
                    {
                        if( XGrabPointer( _glfwWin.Dpy, _glfwWin.Win,
                                True, ButtonPressMask |
                                ButtonReleaseMask | PointerMotionMask,
                                GrabModeAsync, GrabModeAsync,
                                _glfwWin.Win, None, CurrentTime ) ==
                            GrabSuccess )
                        {
                            _glfwWin.PointerGrabbed = GL_TRUE;
                        }
                    }
                }

                // Restore fullscreen mode properties
                if( _glfwWin.Fullscreen )
                {
                    // Change back video mode to user selected mode
                    _glfwSetVideoMode( &_glfwWin.Width, &_glfwWin.Height,
                                       0,0,0 );

                    // Disable window manager decorations
                    wa.override_redirect = True;
                    XChangeWindowAttributes( _glfwWin.Dpy, _glfwWin.Win,
                                             CWOverrideRedirect, &wa );

                    // Make sure window is in upper left corner
                    XMoveWindow( _glfwWin.Dpy, _glfwWin.Win, 0, 0 );
                }

                _glfwWin.Iconified = GL_FALSE;
            }
            break;

        // Was the window activated?
        case FocusIn:
            _glfwWin.Active = GL_TRUE;
            break;

        // Was the window de-activated?
        case FocusOut:
            _glfwWin.Active = GL_FALSE;
            break;

        // Was the window destroyed?
        case DestroyNotify:
            return GL_TRUE;

        default:
            break;
    }

    // The window was not destroyed
    return GL_FALSE;
}



//************************************************************************
//****                    GLFW user functions                         ****
//************************************************************************

//========================================================================
// glfwOpenWindow() - Here is where the window is created, and the OpenGL
// rendering context is created.
//========================================================================

int glfwOpenWindow( int width, int height, int redbits, int greenbits,
    int bluebits, int alphabits, int depthbits, int stencilbits,
    int mode )
{
    Colormap    cmap;
    XSetWindowAttributes wa;
    XEvent      event;
    int         AccumRedBits, AccumGreenBits, AccumBlueBits;
    int         AccumAlphaBits, AuxBuffers, Stereo;

    // Is GLFW initialized?
    if( !_glfwInitialised || _glfwWin.Opened )
    {
        return GL_FALSE;
    }

    // Check input arguments
    if( mode != GLFW_WINDOW && mode != GLFW_FULLSCREEN )
    {
        return GL_FALSE;
    }

    // Clear GLFW state
    _glfwWin.Dpy             = NULL;
    _glfwWin.VI              = NULL;
    _glfwWin.CX              = (GLXContext)0;
    _glfwWin.Win             = (Window)0;
    _glfwWin.Active          = GL_TRUE;
    _glfwWin.Iconified       = GL_FALSE;
    _glfwWin.PointerGrabbed  = GL_FALSE;
    _glfwWin.KeyboardGrabbed = GL_FALSE;
    _glfwClearInput();

    // Unregister all callback functions
    _glfwWindowSizeCallback  = NULL;
    _glfwKeyCallback         = NULL;
    _glfwMousePosCallback    = NULL;
    _glfwMouseButtonCallback = NULL;
    _glfwMouseWheelCallback  = NULL;

    // Get window hints
    AccumRedBits         = _glfwWinHints.AccumRedBits;
    AccumGreenBits       = _glfwWinHints.AccumGreenBits;
    AccumBlueBits        = _glfwWinHints.AccumBlueBits;
    AccumAlphaBits       = _glfwWinHints.AccumAlphaBits;
    AuxBuffers           = _glfwWinHints.AuxBuffers;
    Stereo               = _glfwWinHints.Stereo;

    // Clear window hints
    _glfwWinHints.RefreshRate    = 0;
    _glfwWinHints.AccumRedBits   = 0;
    _glfwWinHints.AccumGreenBits = 0;
    _glfwWinHints.AccumBlueBits  = 0;
    _glfwWinHints.AccumAlphaBits = 0;
    _glfwWinHints.AuxBuffers     = 0;
    _glfwWinHints.Stereo         = 0;

    // Check width & height
    if( width > 0 && height <= 0 )
    {
        // Set the window aspect ratio to 4:3
        height = (width * 3) / 4;
    }
    else if( width <= 0 && height > 0 )
    {
        // Set the window aspect ratio to 4:3
        width = (height * 4) / 3;
    }
    else if( width <= 0 && height <= 0 )
    {
        // Default window size
        width  = 640;
        height = 480;
    }

    // Remeber window settings
    _glfwWin.Width      = width;
    _glfwWin.Height     = height;
    _glfwWin.Fullscreen = (mode == GLFW_FULLSCREEN ? 1 : 0);

    // Fullscreen & screen saver settings
    _glfwFS.ModeChanged     = GL_FALSE;
    _glfwSaver.SaverChanged = GL_FALSE;

    // Get a connection
    _glfwWin.Dpy = XOpenDisplay( 0 );
    if( _glfwWin.Dpy == NULL )
    {
        glfwCloseWindow();
        return GL_FALSE;
    }

    // Check if GLX is supported on this display
    if( !glXQueryExtension( _glfwWin.Dpy, NULL, NULL ) )
    {
        glfwCloseWindow();
        return GL_FALSE;
    }

    // Catch WM close
    _glfwWin.WMDeleteWindow = XInternAtom( _glfwWin.Dpy,
                                           "WM_DELETE_WINDOW",
                                           False );

    // Get an appropriate visual
    _glfwWin.VI = _glfwChooseVisual( _glfwWin.Dpy,
                                     DefaultScreen(_glfwWin.Dpy),
                                     redbits, greenbits, bluebits,
                                     alphabits, depthbits, stencilbits,
                                     AccumRedBits, AccumGreenBits,
                                     AccumBlueBits, AccumAlphaBits,
                                     AuxBuffers, Stereo );
    if( _glfwWin.VI == NULL )
    {
        glfwCloseWindow();
        return GL_FALSE;
    }

    // Create a GLX context
    _glfwWin.CX = glXCreateContext( _glfwWin.Dpy, _glfwWin.VI, 0, GL_TRUE );
    if( _glfwWin.CX == NULL )
    {
        glfwCloseWindow();
        return GL_FALSE;
    }

    // Create a colormap
    cmap = XCreateColormap( _glfwWin.Dpy, RootWindow( _glfwWin.Dpy,
               _glfwWin.VI->screen), _glfwWin.VI->visual, AllocNone );

    // Do we want fullscreen?
    if( mode == GLFW_FULLSCREEN )
    {
        // Change video mode
        _glfwSetVideoMode( &_glfwWin.Width, &_glfwWin.Height,
                           redbits, greenbits, bluebits );

        // Remember old screen saver settings
        XGetScreenSaver( _glfwWin.Dpy, &_glfwSaver.Timeout,
                         &_glfwSaver.Interval, &_glfwSaver.Blanking,
                         &_glfwSaver.Exposure );

        // Disable screen saver
        XSetScreenSaver( _glfwWin.Dpy, 0, 0, DontPreferBlanking,
                         DefaultExposures );
    }

    // Attributes for window
    wa.colormap = cmap;
    wa.border_pixel = 0;
    wa.event_mask = StructureNotifyMask | KeyPressMask | KeyReleaseMask |
        PointerMotionMask | ButtonPressMask | ButtonReleaseMask |
        ExposureMask | FocusChangeMask | VisibilityChangeMask;

    // For fullscreen mode, disable window manager intervention (remove
    // window borders etc). Note: I am aware that this is a fairly ugly
    // brute force way of doing it, but it works...
    wa.override_redirect = (mode == GLFW_FULLSCREEN ? True : False);

    // Create a window
    _glfwWin.Win = XCreateWindow(
        _glfwWin.Dpy,
        RootWindow( _glfwWin.Dpy, _glfwWin.VI->screen ),
        0, 0,                            // Upper left corner
        _glfwWin.Width, _glfwWin.Height, // Width, height
        0,                               // Borderwidth
        _glfwWin.VI->depth,              // Depth
        InputOutput,
        _glfwWin.VI->visual,
        CWBorderPixel | CWColormap | CWEventMask | CWOverrideRedirect,
        &wa
    );
    if( !_glfwWin.Win )
    {
        glfwCloseWindow();
        return GL_FALSE;
    }

    // Allow us to trap the Window Close protocol
    XSetWMProtocols( _glfwWin.Dpy, _glfwWin.Win, &_glfwWin.WMDeleteWindow,
                     1 );

    // Map window
    XMapWindow( _glfwWin.Dpy, _glfwWin.Win );

    // Wait for notification
    XIfEvent( _glfwWin.Dpy, &event, _glfwWaitForNotify,
              (char*)_glfwWin.Win );

    // Make sure that our window ends up on top of things
    XRaiseWindow( _glfwWin.Dpy, _glfwWin.Win );
    XSetInputFocus( _glfwWin.Dpy, _glfwWin.Win, RevertToNone,
                    CurrentTime );

    // Fullscreen mode "post processing"
    if( mode == GLFW_FULLSCREEN )
    {
        // Force window position/size (some WMs do their own window
        // geometry, which we want to override)
        XMoveWindow( _glfwWin.Dpy, _glfwWin.Win, 0, 0 );
        XResizeWindow( _glfwWin.Dpy, _glfwWin.Win, _glfwWin.Width,
                       _glfwWin.Height );

        // Grab keyboard
        if( XGrabKeyboard( _glfwWin.Dpy, _glfwWin.Win, True,
                           GrabModeAsync, GrabModeAsync, CurrentTime ) ==
            GrabSuccess )
        {
            _glfwWin.KeyboardGrabbed = GL_TRUE;
        }

        // Grab mouse cursor
        if( XGrabPointer( _glfwWin.Dpy, _glfwWin.Win, True,
                          ButtonPressMask | ButtonReleaseMask |
                          PointerMotionMask, GrabModeAsync, GrabModeAsync,
                          _glfwWin.Win, None, CurrentTime ) ==
            GrabSuccess )
        {
            _glfwWin.PointerGrabbed = GL_TRUE;
        }

        // Try to get window inside viewport (for virtual displays) by
        // moving the mouse cursor to the upper left corner (and then to
        // the center)
        XWarpPointer( _glfwWin.Dpy, None, _glfwWin.Win, 0,0,0,0, 0,0 );
        XWarpPointer( _glfwWin.Dpy, None, _glfwWin.Win, 0,0,0,0,
                      _glfwWin.Width/2, _glfwWin.Height/2 );
    }

    // Set window & icon name
    XStoreName( _glfwWin.Dpy, _glfwWin.Win, "GLFW Window" );
    XSetIconName( _glfwWin.Dpy, _glfwWin.Win, "GLFW Window" );

    // Connect the context to the window
    glXMakeCurrent( _glfwWin.Dpy, _glfwWin.Win, _glfwWin.CX );

    // Start by clearing the front buffer to black (avoid ugly desktop
    // remains in our OpenGL window)
    //glClear( GL_COLOR_BUFFER_BIT );
    glXSwapBuffers( _glfwWin.Dpy, _glfwWin.Win );

    // Initialize OpenGL extension: GLX_SGI_swap_control
#ifdef GLX_ARB_get_proc_address
    _glXSwapIntervalSGI = (GLXSWAPINTERVALSGI_T)
        glXGetProcAddress( (const GLubyte *) "glXSwapIntervalSGI" );
#else
    _glXSwapIntervalSGI = (GLXSWAPINTERVALSGI_T) 0;
#endif

    // Flag that window is now opened
    _glfwWin.Opened = GL_TRUE;

    // If full-screen mode was requested, lock mouse (hide mouse cursor
    // and enable mouse input)
    if( mode == GLFW_FULLSCREEN )
    {
        glfwDisable( GLFW_MOUSE_CURSOR );
    }

    return GL_TRUE;
}


//========================================================================
// glfwOpenWindowHint() - Set hints for opening the window
//========================================================================

GLFWAPI void GLFWAPIENTRY glfwOpenWindowHint( int target, int hint )
{
    // Is GLFW initialized?
    if( !_glfwInitialised )
    {
        return;
    }

    switch( target )
    {
        case GLFW_REFRESH_RATE:
            _glfwWinHints.RefreshRate = hint;
            break;
        case GLFW_ACCUM_RED_BITS:
            _glfwWinHints.AccumRedBits = hint;
            break;
        case GLFW_ACCUM_GREEN_BITS:
            _glfwWinHints.AccumGreenBits = hint;
            break;
        case GLFW_ACCUM_BLUE_BITS:
            _glfwWinHints.AccumBlueBits = hint;
            break;
        case GLFW_ACCUM_ALPHA_BITS:
            _glfwWinHints.AccumAlphaBits = hint;
            break;
        case GLFW_AUX_BUFFERS:
            _glfwWinHints.AuxBuffers = hint;
            break;
        case GLFW_STEREO:
            _glfwWinHints.Stereo = hint;
            break;
        default:
            break;
    }
}


//========================================================================
// glfwCloseWindow() - Properly kill the video display.
//========================================================================

void glfwCloseWindow( void )
{
    // Is GLFW initialized?
    if( !_glfwInitialised )
    {
        return;
    }

    // Window is no longer opened
    _glfwWin.Opened = GL_FALSE;

    // Unregister all callback functions
    _glfwWindowSizeCallback  = NULL;
    _glfwKeyCallback         = NULL;
    _glfwMousePosCallback    = NULL;
    _glfwMouseButtonCallback = NULL;
    _glfwMouseWheelCallback  = NULL;

    // Show mouse pointer again (if hidden)
    glfwEnable( GLFW_MOUSE_CURSOR );

    // Inactivate OpenGL extension GLX_SGI_swap_control (activated by
    // glfwOpenWindow)
    _glXSwapIntervalSGI = (GLXSWAPINTERVALSGI_T) 0;

    // Do we have a rendering context?
    if( _glfwWin.CX )
    {
        // Release the context
        glXMakeCurrent( _glfwWin.Dpy, None, NULL );

        // Delete the context
        glXDestroyContext( _glfwWin.Dpy, _glfwWin.CX );
        _glfwWin.CX = NULL;
    }

    // Ungrab pointer and/or keyboard?
    if( _glfwWin.KeyboardGrabbed )
    {
        XUngrabKeyboard( _glfwWin.Dpy, CurrentTime );
        _glfwWin.KeyboardGrabbed = GL_FALSE;
    }
    if( _glfwWin.PointerGrabbed )
    {
        XUngrabPointer( _glfwWin.Dpy, CurrentTime );
        _glfwWin.PointerGrabbed = GL_FALSE;
    }

    // Do we have a window?
    if( _glfwWin.Win )
    {
        // Unmap the window
        XUnmapWindow( _glfwWin.Dpy, _glfwWin.Win );

        // Destroy the window
        XDestroyWindow( _glfwWin.Dpy, _glfwWin.Win );
        _glfwWin.Win = (Window) 0;
    }

    // Did we change the fullscreen resolution?
    if( _glfwFS.ModeChanged )
    {
#if defined( GLFW_USE_XF86VIDMODE )
        // Unlock mode switch
        XF86VidModeLockModeSwitch( _glfwWin.Dpy,
                                   DefaultScreen(_glfwWin.Dpy),
                                   0 );

        // Change the video mode back to the old mode
        XF86VidModeSwitchToMode( _glfwWin.Dpy,
            DefaultScreen(_glfwWin.Dpy), &_glfwFS.OldMode );
#endif
        _glfwFS.ModeChanged = GL_FALSE;
    }

    // Did we change the screen saver setting?
    if( _glfwSaver.SaverChanged )
    {
        // Restore old screen saver settings
        XSetScreenSaver( _glfwWin.Dpy, _glfwSaver.Timeout,
                         _glfwSaver.Interval, _glfwSaver.Blanking,
                         _glfwSaver.Exposure );
        _glfwSaver.SaverChanged = GL_FALSE;
    }

    // Do we have a display?
    if( _glfwWin.Dpy )
    {
        // Close display
        XCloseDisplay( _glfwWin.Dpy );
        _glfwWin.Dpy = NULL;
    }
}


//========================================================================
// glfwSetWindowTitle() - Set the window title.
//========================================================================

void glfwSetWindowTitle( const char *title )
{
    // Is GLFW initialized?
    if( !_glfwInitialised || !_glfwWin.Opened )
    {
        return;
    }

    // Set window & icon title
    XStoreName( _glfwWin.Dpy, _glfwWin.Win, title );
    XSetIconName( _glfwWin.Dpy, _glfwWin.Win, title );
}


//========================================================================
// glfwGetWindowSize() - Get the window size.
//========================================================================

void glfwGetWindowSize( int *width, int *height )
{
    // Is GLFW initialized?
    if( !_glfwInitialised || !_glfwWin.Opened )
    {
        return;
    }

    if( width != NULL )
    {
        *width = _glfwWin.Width;
    }
    if( height != NULL )
    {
        *height = _glfwWin.Height;
    }
}


//========================================================================
// glfwSetWindowSize() - Set the window size.
//========================================================================

void glfwSetWindowSize( int width, int height )
{
    int     mode = 0, sizechanged = GL_FALSE;
    GLint   drawbuffer;
    GLfloat clearcolor[4];

    // Is GLFW initialized?
    if( !_glfwInitialised || !_glfwWin.Opened || _glfwWin.Iconified )
    {
        return;
    }

    // If we are in fullscreen mode, get some info about the current mode
    if( _glfwWin.Fullscreen )
    {
        // Get closest match for target video mode
        mode = _glfwGetClosestVideoMode( &width, &height, 0,0,0 );
    }

    // Change window size before changing fullscreen mode?
    if( _glfwWin.Fullscreen && (width > _glfwWin.Width) )
    {
        XResizeWindow( _glfwWin.Dpy, _glfwWin.Win, width, height );
        sizechanged = GL_TRUE;
    }

    // Change fullscreen video mode?
    if( _glfwWin.Fullscreen )
    {
        // Change video mode
        _glfwSetVideoModeMODE( mode );

        // Clear the front buffer to black (avoid ugly desktop remains in
        // our OpenGL window)
        //glGetIntegerv( GL_DRAW_BUFFER, &drawbuffer );
        //glGetFloatv( GL_COLOR_CLEAR_VALUE, clearcolor );
        //glClearColor( 0.0f, 0.0f, 0.0f, 0.0f );
        //glClear( GL_COLOR_BUFFER_BIT );
        if( drawbuffer == GL_BACK )
        {
            glXSwapBuffers( _glfwWin.Dpy, _glfwWin.Win );
        }
        //glClearColor( clearcolor[0], clearcolor[1], clearcolor[2],clearcolor[3] );
    }

    // Set window size (if not already changed)
    if( !sizechanged )
    {
        XResizeWindow( _glfwWin.Dpy, _glfwWin.Win, width, height );
    }
}


//========================================================================
// glfwSetWindowPos() - Set the window position.
//========================================================================

void glfwSetWindowPos( int x, int y )
{
    // Is GLFW initialized?
    if( !_glfwInitialised || !_glfwWin.Opened || _glfwWin.Fullscreen )
    {
        return;
    }

    // Set window position
    XMoveWindow( _glfwWin.Dpy, _glfwWin.Win, x, y );
}


//========================================================================
// glfwIconfyWindow() - Window iconification
//========================================================================

void glfwIconifyWindow( void )
{
    XSetWindowAttributes wa;

    // Is GLFW initialized?
    if( !_glfwInitialised || !_glfwWin.Opened || _glfwWin.Iconified )
    {
        return;
    }

    // In fullscreen mode, we need to restore the desktop video mode
    if( _glfwWin.Fullscreen )
    {
#if defined( GLFW_USE_XF86VIDMODE )
        // Unlock mode switch
        XF86VidModeLockModeSwitch( _glfwWin.Dpy,
                                   DefaultScreen(_glfwWin.Dpy),
                                   0 );

        // Change the video mode back to the old mode
        XF86VidModeSwitchToMode( _glfwWin.Dpy,
            DefaultScreen(_glfwWin.Dpy), &_glfwFS.OldMode );
#endif
        _glfwFS.ModeChanged = GL_FALSE;

        // Allow window manager to iconify the window
        wa.override_redirect = False;
        XChangeWindowAttributes( _glfwWin.Dpy, _glfwWin.Win,
                                 CWOverrideRedirect, &wa );
    }

    // Show mouse pointer
    if( _glfwWin.PointerHidden )
    {
        XUndefineCursor( _glfwWin.Dpy, _glfwWin.Win );
        _glfwWin.PointerHidden = GL_FALSE;
    }

    // Un-grab mouse pointer
    if( _glfwWin.PointerGrabbed )
    {
        XUngrabPointer( _glfwWin.Dpy, CurrentTime );
        _glfwWin.PointerGrabbed = GL_FALSE;
    }

    // Iconify window
    XIconifyWindow( _glfwWin.Dpy, _glfwWin.Win,
                    DefaultScreen(_glfwWin.Dpy) );

    // Window is now iconified
    _glfwWin.Iconified = GL_TRUE;
}


//========================================================================
// glfwRestoreWindow() - Window un-iconification
//========================================================================

void glfwRestoreWindow( void )
{
    XSetWindowAttributes wa;

    // Is GLFW initialized?
    if( !_glfwInitialised || !_glfwWin.Opened || !_glfwWin.Iconified )
    {
        return;
    }

    // In fullscreen mode, change back video mode to user selected mode
    if( _glfwWin.Fullscreen )
    {
        _glfwSetVideoMode( &_glfwWin.Width, &_glfwWin.Height, 0,0,0 );
    }

    // Un-iconify window
    XMapWindow( _glfwWin.Dpy, _glfwWin.Win );

    // In fullscreen mode, remove window border after de-iconification
    if( _glfwWin.Fullscreen )
    {
        // Disable window manager decorations
        wa.override_redirect = True;
        XChangeWindowAttributes( _glfwWin.Dpy, _glfwWin.Win,
                                 CWOverrideRedirect, &wa );

        // Make sure window is in upper left corner
        XMoveWindow( _glfwWin.Dpy, _glfwWin.Win, 0, 0 );
    }

    // Lock mouse, if necessary
    if( _glfwWin.MouseLock )
    {
        // Hide cursor
        if( !_glfwWin.PointerHidden )
        {
            XDefineCursor( _glfwWin.Dpy, _glfwWin.Win,
                           _glfwCreateNULLCursor( _glfwWin.Dpy,
                                                  _glfwWin.Win ) );
            _glfwWin.PointerHidden = GL_TRUE;
        }

        // Grab cursor
        if( !_glfwWin.PointerGrabbed )
        {
            if( XGrabPointer( _glfwWin.Dpy, _glfwWin.Win, True,
                              ButtonPressMask | ButtonReleaseMask |
                              PointerMotionMask, GrabModeAsync,
                              GrabModeAsync, _glfwWin.Win, None,
                              CurrentTime ) == GrabSuccess )
            {
                _glfwWin.PointerGrabbed = GL_TRUE;
            }
        }
    }

    // Window is no longer iconified
    _glfwWin.Iconified = GL_FALSE;
}


//========================================================================
// glfwSwapBuffers() - Swap buffers (double-buffering) and poll any new
// events.
//========================================================================

void glfwSwapBuffers( void )
{
    // Is GLFW initialized?
    if( !_glfwInitialised || !_glfwWin.Opened )
    {
        return;
    }

    // Check for window messages
    glfwPollEvents();

    // Update display-buffer
    if( _glfwWin.Opened )
    {
        glXSwapBuffers( _glfwWin.Dpy, _glfwWin.Win );
    }
}


//========================================================================
// glfwSwapInterval() - Set double buffering swap interval (0 = vsync off)
//========================================================================

void glfwSwapInterval( int interval )
{
    // Is GLFW initialized?
    if( !_glfwInitialised || !_glfwWin.Opened )
    {
        return;
    }

    if( _glXSwapIntervalSGI )
    {
        _glXSwapIntervalSGI( interval );
    }
}


//========================================================================
// glfwGetWindowParam()
//========================================================================

int glfwGetWindowParam( int param )
{
    int value;
#if defined( GLFW_USE_XF86VIDMODE )
    XF86VidModeModeLine modeline;
    int dotclock;
    float pixels_per_second, pixels_per_frame;
#endif

    // Is GLFW initialized?
    if( !_glfwInitialised )
    {
        return 0;
    }

    // Is the window opened?
    if( !_glfwWin.Opened )
    {
        if( param == GLFW_OPENED )
        {
            return GL_FALSE;
        }
        return 0;
    }

    // Get the requested format parameter
    switch( param )
    {
    case GLFW_OPENED:
        value = GL_TRUE;
        break;
    case GLFW_ACTIVE:
        value = _glfwWin.Active;
        break;
    case GLFW_ICONIFIED:
        value = _glfwWin.Iconified;
        break;
    case GLFW_ACCELERATED:
        value = GL_TRUE; // AFAIK, there is no easy/sure way of knowing...
        break;
    case GLFW_RED_BITS:
        glXGetConfig( _glfwWin.Dpy, _glfwWin.VI, GLX_RED_SIZE, &value );
        break;
    case GLFW_GREEN_BITS:
        glXGetConfig( _glfwWin.Dpy, _glfwWin.VI, GLX_GREEN_SIZE, &value );
        break;
    case GLFW_BLUE_BITS:
        glXGetConfig( _glfwWin.Dpy, _glfwWin.VI, GLX_BLUE_SIZE, &value );
        break;
    case GLFW_ALPHA_BITS:
        glXGetConfig( _glfwWin.Dpy, _glfwWin.VI, GLX_ALPHA_SIZE, &value );
        break;
    case GLFW_DEPTH_BITS:
        glXGetConfig( _glfwWin.Dpy, _glfwWin.VI, GLX_DEPTH_SIZE, &value );
        break;
    case GLFW_STENCIL_BITS:
        glXGetConfig( _glfwWin.Dpy, _glfwWin.VI, GLX_STENCIL_SIZE,
                      &value );
        break;
    case GLFW_ACCUM_RED_BITS:
        glXGetConfig( _glfwWin.Dpy, _glfwWin.VI, GLX_ACCUM_RED_SIZE,
                      &value );
        break;
    case GLFW_ACCUM_GREEN_BITS:
        glXGetConfig( _glfwWin.Dpy, _glfwWin.VI, GLX_ACCUM_GREEN_SIZE,
                      &value );
        break;
    case GLFW_ACCUM_BLUE_BITS:
        glXGetConfig( _glfwWin.Dpy, _glfwWin.VI, GLX_ACCUM_BLUE_SIZE,
                      &value );
        break;
    case GLFW_ACCUM_ALPHA_BITS:
        glXGetConfig( _glfwWin.Dpy, _glfwWin.VI, GLX_ACCUM_ALPHA_SIZE,
                      &value );
        break;
    case GLFW_AUX_BUFFERS:
        glXGetConfig( _glfwWin.Dpy, _glfwWin.VI, GLX_AUX_BUFFERS,
                      &value );
        break;
    case GLFW_STEREO:
        glXGetConfig( _glfwWin.Dpy, _glfwWin.VI, GLX_STEREO, &value );
        value = value ? GL_TRUE : GL_FALSE;
        break;
    case GLFW_REFRESH_RATE:
#if defined( GLFW_USE_XF86VIDMODE )
        // Use the XF86VidMode extension to get current video mode
        XF86VidModeGetModeLine( _glfwWin.Dpy,
            DefaultScreen(_glfwWin.Dpy), &dotclock, &modeline );
        pixels_per_second = 1000.0f * (float) dotclock;
        pixels_per_frame  = (float) modeline.htotal * modeline.vtotal;
        value = (int) (pixels_per_second / pixels_per_frame + 0.5f);
#else
        value = 0;
#endif
        break;
    default:
        value = 0;
        break;
    }

    return value;
}


//========================================================================
// glfwSetWindowSizeCallback() - Set callback function for window size
// changes
//========================================================================

void glfwSetWindowSizeCallback( GLFWwindowsizefun cbfun )
{
    // Is GLFW initialized?
    if( !_glfwInitialised || !_glfwWin.Opened )
    {
        return;
    }

    // Use default callback?
    _glfwWindowSizeCallback = cbfun;

    // Call the callback function to let the application know the current
    // window size
    if( _glfwWindowSizeCallback )
    {
        _glfwWindowSizeCallback( _glfwWin.Width, _glfwWin.Height );
    }
}


//========================================================================
// glfwPollEvents() - Poll for new window and input events
//========================================================================

void glfwPollEvents( void )
{
    int winclosed;

    // Is GLFW initialized?
    if( !_glfwInitialised || !_glfwWin.Opened )
    {
        return;
    }

    // Flag: mouse was not moved (will be changed by _glfwGetNextEvent if
    // there was a mouse move event)
    _glfwInput.MouseMoved = GL_FALSE;
    if( _glfwWin.MouseLock )
    {
        _glfwInput.OldMouseX = _glfwWin.Width/2;
        _glfwInput.OldMouseY = _glfwWin.Height/2;
    }
    else
    {
        _glfwInput.OldMouseX = _glfwInput.MousePosX;
        _glfwInput.OldMouseY = _glfwInput.MousePosY;
    }

    // Use XSync to synchronise events to the X display.
    // I don't know if this can have a serious performance impact. My
    // benchmarks with a GeForce card under Linux shows no difference with
    // or without XSync, but when the GL window is rendered over a slow
    // network I have noticed bad event syncronisation problems when XSync
    // is not used, so I decided to use it.
    XSync( _glfwWin.Dpy, False );

    // Get next window event
    winclosed = GL_FALSE;
    while( XPending( _glfwWin.Dpy ) )
    {
        if( _glfwGetNextEvent() )
        {
            winclosed = GL_TRUE;
        }
    }

    // Did we have mouse movement in locked cursor mode?
    if( _glfwInput.MouseMoved && _glfwWin.MouseLock )
    {
        // Move the mouse pointer back to the window centre so that it
        // does not wander off...
        XWarpPointer( _glfwWin.Dpy, None, _glfwWin.Win, 0,0,0,0,
                      _glfwWin.Width/2, _glfwWin.Height/2 );
    }

    // Was there a window close request?
    if( winclosed )
    {
        glfwCloseWindow();
    }
}
