open Gl

(* shortcut for int Bigarray *)
type tbaint = (int, Bigarray.int_elt, Bigarray.c_layout) Bigarray.Array1.t

(* shortcut for float Bigarray *)
type tbafloat = (float, Bigarray.float32_elt, Bigarray.c_layout) Bigarray.Array1.t

(* shortcut for double Bigarray *)
type tbadouble = (float, Bigarray.float64_elt, Bigarray.c_layout) Bigarray.Array1.t

(* shortcut for char Bigarray *)
type tbabyte = (int, Bigarray.int8_unsigned_elt, Bigarray.c_layout) Bigarray.Array1.t

let ba_int l = Bigarray.Array1.create Bigarray.int Bigarray.c_layout l
let ba_float l = Bigarray.Array1.create Bigarray.float32 Bigarray.c_layout l
let ba_double l = Bigarray.Array1.create Bigarray.float64 Bigarray.c_layout l
let ba_byte l = Bigarray.Array1.create Bigarray.int8_unsigned Bigarray.c_layout l

external ba_int_zero: int -> (int, Bigarray.int_elt, Bigarray.c_layout) Bigarray.Array1.t = "c_proml_types_int_zero"
external ba_float_zero: int -> (float, Bigarray.float32_elt, Bigarray.c_layout) Bigarray.Array1.t = "c_proml_types_float_zero"
external ba_double_zero: int ->(float, Bigarray.float64_elt, Bigarray.c_layout) Bigarray.Array1.t = "c_proml_types_double_zero"
external ba_byte_zero: int -> (int, Bigarray.int8_unsigned_elt, Bigarray.c_layout) Bigarray.Array1.t = "c_proml_types_byte_zero"

external ba_int_set: int -> int -> (int, Bigarray.int_elt, Bigarray.c_layout) Bigarray.Array1.t = "c_proml_types_int_set"
external ba_float_set: int -> float -> (float, Bigarray.float32_elt, Bigarray.c_layout) Bigarray.Array1.t = "c_proml_types_float_set"
external ba_double_set: int -> float -> (float, Bigarray.float64_elt, Bigarray.c_layout) Bigarray.Array1.t = "c_proml_types_double_set"
external ba_byte_set: int -> int -> (int, Bigarray.int8_unsigned_elt, Bigarray.c_layout) Bigarray.Array1.t = "c_proml_types_byte_set"

external ba_byte_rgb: int -> int -> int -> int -> (int, Bigarray.int8_unsigned_elt, Bigarray.c_layout) Bigarray.Array1.t = "c_proml_types_byte_rgb"
external ba_byte_rgba: int -> int -> int -> int -> int -> (int, Bigarray.int8_unsigned_elt, Bigarray.c_layout) Bigarray.Array1.t = "c_proml_types_byte_rgba"

external ba_float_init: float array -> tbafloat = "c_proml_types_float_init"
external ba_double_init: float array -> tbadouble = "c_proml_types_double_init"
external ba_int_init: int array -> tbaint = "c_proml_types_int_init"
external ba_byte_init: int array -> tbabyte = "c_proml_types_byte_init"

external ba_ofs_int: ('a, 'b, Bigarray.c_layout) Bigarray.Array1.t -> int -> int -> tbaint = "ml__baofs_int"
external ba_ofs_float: ('a, 'b, Bigarray.c_layout) Bigarray.Array1.t -> int -> int -> tbafloat = "ml__baofs_float"
external ba_ofs_double: ('a, 'b, Bigarray.c_layout) Bigarray.Array1.t -> int -> int -> tbadouble = "ml__baofs_double"
external ba_ofs_byte: ('a, 'b, Bigarray.c_layout) Bigarray.Array1.t -> int -> int -> tbabyte = "ml__baofs_uint8"

let ba_size = Bigarray.Array1.dim 
let ba_length = Bigarray.Array1.dim 

let ba_iteri f ba =
    for i = 0 to ba_size ba - 1 do
	f i ba.{i}
    done

let ba_iter f ba =
    for i = 0 to ba_size ba - 1 do
	f ba.{i}
    done

let get_integer state =
    let a = ba_int 1 in
    glGetIntegerv state a;
    a.{0}
    
let get_boolean state =
    let a = ba_byte 1 in
    glGetBooleanv state a;
    if a.{0} = 0 then false else true
    
let get_float state =
    let a = ba_float 1 in
    glGetFloatv state a;
    a.{0}
    
let get_double state =
    let a = ba_double 1 in
    glGetDoublev state a;
    a.{0}
    
let test_glerror s =
	let e = glGetError () in
	let se () =
		if e = cgl_no_error then "No Error"
		else if e = cgl_invalid_value then "Invalid Value"
		else if e = cgl_invalid_enum then "Invalid Enum"
		else if e = cgl_invalid_operation then "Invalid Operation"
		else if e = cgl_stack_overflow then "Stack Overflow"
		else if e = cgl_stack_underflow then "Stack Underflow"
		else if e = cgl_out_of_memory then "Out of Memory"
		else if e = cgl_table_too_large then "Table too Large"
		else "Unknown Error: "^(string_of_int e)
	in
	if e <> cgl_no_error then begin
		Printf.eprintf "GL_ERROR (%s): %s\n" s (se ()); 
		flush stdout;
	end

let make_list f =
	let i = glGenLists 1 in
	glNewList i cgl_compile;
	f ();
	glEndList ();
	i

let new_texture () =
	let a = ba_int 1 in
	glGenTextures 1 a;
	a.{0}
	
let del_texture i =
	let a = ba_int 1 in
	a.{0} <- i;
	glDeleteTextures 1 a
